//    FILE: I2C_eeprom.cpp
//  AUTHOR: Rob Tillaart
// VERSION: 0.2.00
// PURPOSE: Simple I2C_eeprom library for Arduino with EEPROM 24LC256 et al.
//
// HISTORY: 
// 0.1.00 - 2011-01-21 initial version
// 0.1.01 - 2011-02-07 added setBlock function
// 0.2.00 - 2011-02-11 fixed 64 bit boundary bug
// 
// Released to the public domain
//

#include <I2C_eeprom.h>

#include "Wstring.h"
#include "Wiring.h"
#include <HardwareSerial.h>

////////////////////////////////////////////////////////////////////
//
// PUBLIC
//

I2C_eeprom::I2C_eeprom(uint8_t device, uint32_t space)
{
	_Device = device;
	chip_space = space;
	Wire.begin(); 		// initialise the connection
}

void I2C_eeprom::writeByte(unsigned int address, uint8_t data )
{
	_WriteBlock(address, &data, 1);
}

void I2C_eeprom::writeInt(unsigned int address, uint16_t data )
{
	uint8_t buff[2];
	buff[1]  = (data & 0xFF);
    buff[0] = ((data >> 8) & 0xFF);      
	writeByte( address, buff[0]);
	writeByte( address+1, buff[1]);
	// Does not work like this... Why? :s
	//	_WriteBlock(address, buff, 2);
}

void I2C_eeprom::writeBlock(unsigned int address, uint8_t* buffer, int length)
{
	// determine length until end of page
	int le = endOfPage(address);
	if (le > 0)
	{
		_WriteBlock(address, buffer, le);
		address += le;
		buffer += le;
		length -= le;
	}

	// write the rest at BLOCKSIZE (16) byte boundaries 
	while (length > 0)
	{
		_WriteBlock(address, buffer, min(length, BLOCKSIZE));
		address += BLOCKSIZE;
		buffer += BLOCKSIZE;
		length -= BLOCKSIZE;
	}
}

void I2C_eeprom::setBlock(unsigned int address, uint8_t data, int length)
{
	uint8_t buffer[BLOCKSIZE];
	for (uint8_t i =0; i< BLOCKSIZE; i++) buffer[i] = data;

	// determine length until end of page
	int le = endOfPage(address);
	if (le > 0)
	{
		_WriteBlock(address, buffer, le);
		address += le;
		length -= le;
	}

	while (length > 0)
	{
		_WriteBlock(address, buffer, min(length, BLOCKSIZE));
		address += BLOCKSIZE;
		length -= BLOCKSIZE;
	}
}

uint8_t I2C_eeprom::readByte(unsigned int address)
{
	uint8_t rdata;
	_ReadBlock(address, &rdata, 1);
	return rdata;
}

uint16_t I2C_eeprom::readInt(unsigned int address)
{
	uint8_t buff[2];
	_ReadBlock(address, buff, 2);
	return (buff[0] << 8) | buff[1] ;
}
/*
// maybe let's not read more than 30 or 32 uint8_ts at a time!
void I2C_eeprom::readBlock(unsigned int address, uint8_t* buffer, int length)
{
	while (length > 0)
	{
		_ReadBlock(address, buffer, min(length, BLOCKSIZE));
		address += BLOCKSIZE;
		buffer += BLOCKSIZE;
		length -= BLOCKSIZE;
	}
}
*/

void I2C_eeprom::readBlock(unsigned int address, uint8_t* buffer, int length)
{
	// determine length until end of page
	int le = endOfPage(address);
	if (le > 0)
	{
		_ReadBlock(address, buffer, le);
		address += le;
		buffer += le;
		length -= le;
	}

	// write the rest at BLOCKSIZE (16) byte boundaries 
	while (length > 0)
	{
		_ReadBlock(address, buffer, min(length, BLOCKSIZE));
		address += BLOCKSIZE;
		buffer += BLOCKSIZE;
		length -= BLOCKSIZE;
	}

}

////////////////////////////////////////////////////////////////////
//
// PRIVATE
//

// detemines length until first multiple of 16 of an address
// so writing allways occurs up to 16 byte boundaries
// this is automatically 64 byte boundaries
int I2C_eeprom::endOfPage(unsigned int address)
{
	const int m = BLOCKSIZE;
	unsigned int eopAddr = ((address + m - 1) / m) * m;  // "end of page" address
	return eopAddr - address;  // length until end of page
}

// pre: length < 32;
void I2C_eeprom::_WriteBlock(unsigned int address, uint8_t* buffer, uint8_t length)
{
	int chip_i2c_address = _Device+(address/chip_space);
	int chip_data_address = address%chip_space;
	Wire.beginTransmission(chip_i2c_address);
	Wire.send((int)(chip_data_address >> 8)); 
	Wire.send((int)(chip_data_address & 0xFF));  
	for (uint8_t c = 0; c < length; c++)
	Wire.send(buffer[c]);
	Wire.endTransmission();
	delay(5);
}

// pre: buffer is large enough to hold length bytes
void I2C_eeprom::_ReadBlock(unsigned int address, uint8_t* buffer, uint8_t length)
{
	byte chip_i2c_address = _Device+(address/chip_space);
	int chip_data_address = address%chip_space;
	Wire.beginTransmission(chip_i2c_address);
	Wire.send((int)(chip_data_address >> 8));
	Wire.send((int)(chip_data_address & 0xFF));
	Wire.endTransmission();
	Wire.requestFrom(chip_i2c_address, length);
	for (int c = 0; c < length; c++ )
	if (Wire.available()) buffer[c] = Wire.receive();
}

void I2C_eeprom::dumpEEPROM(unsigned int addr, unsigned int length)
{
  // blocks of 16 bytes
  addr = addr / BLOCKSIZE * BLOCKSIZE;
  length = (length + 9)/BLOCKSIZE * BLOCKSIZE;

  byte b = readByte(addr); 
  for (int i = 0; i < length; i++) 
  {
    if (addr % BLOCKSIZE == 0)
    {
      Serial.println();
      Serial.print(addr);
      Serial.print(":\t");
    }
	if (b<0x10) Serial.print('0');
    Serial.print(b, HEX);
    b = readByte(++addr); 
    Serial.print("  ");
  }
  Serial.println();
}

//
// END OF FILE
//
